/**
 * Project: led line with max7219 driver
 * Author: ruX [Ruslan Zaharov] http://ruX.pp.ru/
 * Date: 25.03.2010
 */

#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>


#define PIN_CS1 4
#define PIN_CS2 5

// Clock duration
#define CLOCK_TIME_MS 0.0005

#define LOW 0
#define HIGH 1

static unsigned char redMatrix[] = {4, 2, 0, 6}, greenMatrix[] = {6, 4, 2, 0};

enum LedStatus{
    lsOff, lsGreen, lsRed, lsOrange
};


char buffman_changed = 0;
char buffman_data[13];

LedStatus buffman_get(unsigned int num) {
    if (num > 52) return lsOff;
    int bitcount = (num * 2);
    int off_bit = bitcount % 8;
    return (LedStatus) ((buffman_data[bitcount/8] & (3 << off_bit)) >> off_bit);
}
void buffman_setAll(LedStatus stat) {
    unsigned char mask = stat << 0 | stat << 2 | stat << 4 | stat << 6;
    for(int i = 0; i < 13; i++) buffman_data[i] = mask;
    buffman_changed = 1;
}
void buffman_set(unsigned char num, LedStatus stat) {
    if (num > 51) return;
    int bitcount = (num * 2);
    int off_byte = num >> 2, off_bit = bitcount % 8;
    buffman_data[off_byte] = (buffman_data[off_byte] & ~(3 << off_bit)) | (((char)stat) << off_bit);
    buffman_changed = 1;
}


void led_cs1(char x) {
    PORTC = x ? PORTC | _BV(4) : PORTC & ~_BV(4);
}
void led_cs2(char x) {
    PORTC = x ? PORTC | _BV(5) : PORTC & ~_BV(5);
}
void led_clk(char x) {
    PORTD = x ? PORTD | _BV(7) : PORTD & ~_BV(7);
}
void led_din(char x) {
    PORTD = x ? PORTD | _BV(6) : PORTD & ~_BV(6);
}
//#define led_din(x) PORTD = x ? PORTD | _BV(6) : PORTD & ~_BV(6);


void outputByte(unsigned char val) {
    for(signed char i = 7; i >= 0; i--) {
        led_din(val & (1 << i));
        _delay_ms(CLOCK_TIME_MS /2);
        
        led_clk(HIGH);
        _delay_ms(CLOCK_TIME_MS);
        
        led_clk(LOW);
        _delay_ms(CLOCK_TIME_MS);
    }
}

void outBytes(char pin, unsigned char reg, unsigned char dat) {
    PORTC &= ~_BV(pin);
    outputByte(reg);
    outputByte(dat);
    PORTC |= _BV(pin);
}

#define outBytesChip1(reg, dat) outBytes(PIN_CS1, reg, dat);
#define outBytesChip2(reg, dat) outBytes(PIN_CS2, reg, dat);
        
void outBytesChips(unsigned char reg, unsigned char dat) {
    outBytesChip1(reg, dat);
    outBytesChip2(reg, dat);
}

void init() {
    // configure pins to output
    DDRC |= _BV(5) | _BV(4);
    DDRD |= _BV(6) | _BV(7);
    
    // pre-init drivers line
    led_cs1(HIGH);
    led_cs2(HIGH);
    led_clk(LOW);

    // setup clock interrupt and prescaler
    TCCR0 |= (1<<CS02)|(1<<CS00);
    TIMSK |= (1<<TOIE0);
    TCNT0 = 0;
    
    // initialize drivers
    outBytesChips(0xc, (1)); // shutdown
    outBytesChips(0xa, (15)); // intensity
    outBytesChips(0x9, 0); // decode mode
    outBytesChips(0xf, 0); // test mode
    outBytesChips(0xb, 7); // scan limit
    
    // finally, enable interrupts
    sei();
}

ISR(TIMER0_OVF_vect) { 
    int mask = 0;
    if (!buffman_changed) return; 
        
    for(int i = 0; i < 15; i++) {
        mask = 0;
        for(int j = 0; j < 4; j++) {
            switch(buffman_get(i*4+j)) {
            case lsOff: continue;
            case lsOrange: mask |= 2 << redMatrix[j] | 1 << greenMatrix[j]; break;
            case lsRed: mask |= 2 << redMatrix[j]; break;
            case lsGreen: mask |= 1 << greenMatrix[j]; break;
            }
        }
        
        if (i < 8) {
            outBytesChip2(i+1, mask);
        } else {
            outBytesChip1(i-7, mask);
        }
    }
}

inline void wait(float ms) {
    for(; ms > 0; ms-=1) _delay_ms(1);
}

int main(void)
{
    init();
    
    buffman_setAll(lsOff);

    int dir = 1, i = 1;
    LedStatus ls = lsRed;
    do {
        buffman_set(i-1, lsOrange);
        buffman_set(i, lsRed);
        buffman_set(i+1, lsRed);
        buffman_set(i+2, lsGreen);
        if (i == 0 || i == 49) dir = - dir;
        i+=dir;
        wait(25);
        if (i%2) buffman_set(0, (buffman_get(0) == lsOff ? lsRed : lsOff ));
        if (i%2==0) buffman_set(51, (buffman_get(51) == lsOff ? lsRed : lsOff ));
        
    } while( 1 );
    

    return 0;
}
